/*!
   memfifo -- Connects the bi-directional high speed interface of default firmware to a FIFO built of on-board SDRAM or on-chip BRAM
   Copyright (C) 2009-2014 ZTEX GmbH.
   http://www.ztex.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!*/

import java.io.*;
import java.util.*;
import java.text.*;

import ch.ntb.usb.*;

import ztex.*;

// *****************************************************************************
// ******* ParameterException **************************************************
// *****************************************************************************
// Exception the prints a help message
class ParameterException extends Exception {
    public final static String helpMsg = new String (
		"Parameters:\n"+
		"    -d <number>       Device Number (default: 0)\n" +
		"    -r 	       Reset EZ-USB\n" +
		"    -v 	       Print debug info\n" +
		"    -p                Print bus info\n" +
		"    -h                This help" );
    
    public ParameterException (String msg) {
	super( msg + "\n" + helpMsg );
    }
}

// *****************************************************************************
// ******* USBReader ***********************************************************
// *****************************************************************************
class UsbReader extends Thread {
    private final int bufNum = 8;
    public final int bufSize = 512*1024;
    public byte[][] buf = new byte[bufNum][];
    public int[] bufBytes = new int[bufNum];
    private int readCount = -1;
    private int getCount = -1;
    private boolean terminate = false;
    private Ztex1v1 ztex;
    public String errorString = "(No error)";
    private long byteCount = 0;
    public UsbWriter writer = null;			// only read if enough data is written in order to prevent dead locks on certain USB implementations
    
    public UsbReader ( Ztex1v1 p_ztex ) {
        super ();
        ztex = p_ztex;
	for (int i=0; i<bufNum; i++) {
	    buf[i]=new byte[bufSize];
	}
    }
    
    public int getBuffer () {
	getCount += 1;
	while ( (getCount >= readCount) && !terminate ) {
	    try { sleep(2); } catch ( InterruptedException e) { } 
	}
	return getCount % bufNum;
    }

    public void reset () {
	getCount = readCount + 1;
	byteCount = 0;
	errorString = "(No error)";
    }

    public long byteCount () {
	return byteCount;
    } 

    public void terminate() {
	terminate = true;
    }

    public void run() {
	setPriority(MAX_PRIORITY);

// reader loop
	while ( !terminate ) {
	    readCount += 1;
	    
	    while ( (readCount-bufNum>=getCount) || ( (writer!=null) && (writer.byteCount()<byteCount+bufSize) ) ) {
		try { sleep(1); } catch ( InterruptedException e) { } 
//		yield();
        	if ( terminate ) return;
	    }

	    int i = readCount % bufNum;
	    int j = 5;
	    String s = null;
	    do {	
		try {
		    bufBytes[i] = LibusbJava.usb_bulk_read(ztex.handle(), ztex.defaultInEP(), buf[i], bufSize, 1000);
		}
		catch (Exception e) {
	    	    s = e.getLocalizedMessage();
	    	    bufBytes[i] = 0;
		}
		j--;
	    } while ( !terminate && (bufBytes[i]<=0) && (j>0) );  // also accept short packages

	    if ( bufBytes[i]<=0 ) {
		errorString = s == null ? "Unknown error " + bufBytes[i] : s;
	    } else {
		byteCount += bufBytes[i];
	    }
//	    System.out.println("Buffer " + i +": read " + bufBytes[i] + " bytes");
	}

    }
}

// *****************************************************************************
// ******* UsbWriter ***********************************************************
// *****************************************************************************
class UsbWriter extends Thread {
    private final int bufNum = 8;
    public final int bufSize = 512*1024;
    public byte[][] buf = new byte[bufNum][];
    private Ztex1v1 ztex;
    private int writeCount = -1;
    private int getCount = -1;
    private boolean terminate = false;
    public String errorString = null;		// null means no error
    private long byteCount = 0;

    public UsbWriter ( Ztex1v1 p_ztex ) {
        super ();
        ztex = p_ztex;
	for (int i=0; i<bufNum; i++) {
	    buf[i]=new byte[bufSize];
	}
    }
    
    public int getBuffer () {
	getCount += 1;
	while ( (getCount - bufNum >= writeCount) && !terminate ) {
	    try { sleep(2); } catch ( InterruptedException e) { } 
	}
	return getCount % bufNum;
    }

    public void reset () {
	getCount = writeCount + 1;
	errorString = null;
	byteCount = 0;
    }
    
    public long byteCount () {
	return byteCount;
    } 

    public void terminate() {
	terminate = true;
    }

    public void run() {
	setPriority(MAX_PRIORITY);

// writer loop
	while ( !terminate ) {
	    writeCount += 1;
	    
	    while ( (writeCount >= getCount) ) {
		try { sleep(1); } catch ( InterruptedException e) { } 
//		yield();
        	if ( terminate ) return;
	    }

	    int i = writeCount % bufNum;
	    int j;
	    int k=5;
	    String s = null;
	    do {
		try {
		    j = LibusbJava.usb_bulk_write(ztex.handle(), ztex.defaultOutEP(), buf[i], bufSize, 1000);
		}
		catch ( Exception e ) {
	    	    s = e.getLocalizedMessage();
		    j = 0;
		}
		k--;
	    } while ( !terminate && (j<=0) && (k>0) );
	    
	    if ( !terminate && (j!=bufSize) ) {
		errorString = s != null ? s : j < 0 ? "Error " + j : "Unknown error: wrote " + j  + " / " + bufSize + " bytes";
	    }

	    byteCount += Math.max(0,j);
	}
	    
    }
}

// *****************************************************************************
// ******* UsbTestWriter *******************************************************
// *****************************************************************************
class UsbTestWriter extends Thread {
    public boolean terminate = false;
    private UsbWriter writer;

    public UsbTestWriter ( Ztex1v1 p_ztex ) {
        super ();
	writer = new UsbWriter( p_ztex );
    }
    
    public UsbWriter writer() {
	return writer;
    }

    public void terminate() {
	terminate = true;
	writer.terminate();
    }

    public void run() {
	writer.start();

	int k = 0;
	int cs = 47;
	int sync;
	Random random = new Random();
	while ( !terminate ) {
	    byte[] b = writer.buf[writer.getBuffer()];
	    for ( int i=0; !terminate && i<writer.bufSize; i++ ) {
		int j = k & 15;
		sync = ( ((j & 1)==1) || (j==14) ) ? 128 : 0;
		if ( j == 15 ) {
		    b[i] = (byte) (sync | ((cs & 127) ^ (cs>>7)));
		    cs = 47;
		}
		else {
//		    b[i] = (byte) ( (j==0 ? (k>>4) & 127 : random.nextInt(128)) | sync );
		    b[i] = (byte) ( (((k>>4)+j) & 127) | sync );
		    cs += (b[i] & 255);
		}
		k=(k+1) & 65535;
	    }
	}
	
	while ( writer.isAlive() ) {
	    try {
		Thread.sleep(1);
            }
    	    catch ( InterruptedException e) {
            } 
	}
    }
}


// *****************************************************************************
// ******* MemFifo *************************************************************
// *****************************************************************************
class MemFifo extends Ztex1v1 {

    // constructor
    public MemFifo ( ZtexDevice1 pDev ) throws UsbException {
	super ( pDev );
    }

    // set mode
    public void setMode( int i ) throws InvalidFirmwareException, UsbException, CapabilityException {
	defaultGpioCtl(3,i);
    }

    // reset
    public void reset( ) throws  InvalidFirmwareException, UsbException, CapabilityException {
	defaultGpioCtl(3,0);
	defaultReset();
	debug2PrintNextLogMessages(System.out); 
	byte[] buf = new byte[4096];
	while ( LibusbJava.usb_bulk_read(handle(), defaultInEP(), buf, buf.length, 100) == buf.length ) { };  // empties buffers
    }

    // reads data and verifies them. 
    // returns true if errors occurred
    // rate is data rate in kBytes; <=0 means unlimited
    public boolean verify ( UsbReader reader, int iz, int rate, boolean verbose) {
    	boolean valid = false;
	int byte_cnt = 0, sync_cnt = 0, cs = 47;
	boolean memError = false;

	for (int i=0; i<iz; i++) {
	    if ( i>0 && rate>0 ) {
		try {
		    Thread.sleep(reader.bufSize/rate);
        	}
        	catch ( InterruptedException e) {
        	} 
	    }
	    int j = reader.getBuffer();
	    int bb = reader.bufBytes[j];
	    byte[] b = reader.buf[j];
	    int serrors = 0;
	    int merrors = 0;

	    if ( bb <= 0 ) {
		System.out.println("Data read error: "+reader.errorString);
		return true;
	    }

	    for (int k=0; k<bb; k++ ) {
		byte_cnt++;
		sync_cnt++;
		if ( (b[k] & 128) == 0 ) sync_cnt=0;
		if ( sync_cnt == 3 ) {
		    boolean serror = byte_cnt != 16;
		    boolean merror = (byte_cnt == 16) && ( ( b[k] & 127 ) != ((cs & 127) ^ (cs>>7)) );
		    valid = valid || ( !serror && !merror );
		    if ( valid ) {
			if ( serror ) serrors += 1;
			if ( merror ) merrors += 1;
			if ( verbose && ((serror && serrors <= 2) || (merror && merrors <= 2) )) {
			    if ( serror ) System.out.print( "Sync Error: " + byte_cnt + " at " + k + ": " );
			    else System.out.print( "Data Error: at " + k + ": " );
			    for ( int l=k-byte_cnt-2; l<k+3; l++ )
				if ( (l>=0) && (l<bb) ) 
				    System.out.print((b[l] < 0 ? "*" : "" ) + (b[l] & 127) + " ");
			}
		    }
		    cs=47;
		    byte_cnt = 0;
		} else {
		    cs += b[k] & 255;
		}
	    }
		    
	    if ( ! valid ) {
		System.out.println("Invalid data");
		return true;
	    }
		    
	    System.out.print("Buffer " + i + (bb==reader.bufSize ? ": " : "(short packet): ") + serrors + " sync errors, 	" + merrors + " memory or transfer errors" + 
		(reader.writer!=null && reader.writer.errorString!=null ? ", 	write errors occurred: " + reader.writer.errorString : "" ) + "          \r" );
	    if ( merrors + serrors > (i==0 ? 2 : 0 ) ) { // two error in first buffer may occurs after changing mode
	        System.out.println();
	        memError |= (bb!=reader.bufSize);
	    } else if (bb!=reader.bufSize) {
	        System.out.println();
	    }
	}
        System.out.println();
	return memError;
    }

// ******* main ****************************************************************
    public static void main (String args[]) {
    
	int devNum = 0;
	boolean verbose = false;
	boolean reset = false;
	UsbReader reader = null;
	UsbTestWriter writer = null;
	
	try {
// init USB stuff
	    LibusbJava.usb_init();

// scan the USB bus
	    ZtexScanBus1 bus = new ZtexScanBus1( ZtexDevice1.ztexVendorId, ZtexDevice1.ztexProductId, true, false, 1);
	    if ( bus.numberOfDevices() <= 0) {
		System.err.println("No devices found");
	        System.exit(0);
	    }
	    
// scan the command line arguments
    	    for (int i=0; i<args.length; i++ ) {
	        if ( args[i].equals("-d") ) {
	    	    i++;
		    try {
			if (i>=args.length) throw new Exception();
    			devNum = Integer.parseInt( args[i] );
		    } 
		    catch (Exception e) {
		        throw new ParameterException("Device number expected after -d");
		    }
		}
		else if ( args[i].equals("-v") ) {
		    verbose = true;
		}
		else if ( args[i].equals("-r") ) {
		    reset = true;
		}
		else if ( args[i].equals("-p") ) {
	    	    bus.printBus(System.out);
		    System.exit(0);
		}
		else if ( args[i].equals("-p") ) {
	    	    bus.printBus(System.out);
		    System.exit(0);
		}
		else if ( args[i].equals("-h") ) {
		        System.err.println(ParameterException.helpMsg);
	    	        System.exit(0);
		}
		else throw new ParameterException("Invalid Parameter: "+args[i]);
	    }

	    String errStr = "";

// create the main class	    
	    MemFifo ztex = new MemFifo ( bus.device(devNum) );
	    
	    try {
	    
// reset EZ-USB
		if ( reset ) {
		    System.out.println("Reset EZ-USB");
		    ztex.resetEzUsb();
		}

// check for firmware
		ztex.defaultCheckVersion(1);
		ztex.debug2PrintNextLogMessages(System.out); 
		if (ztex.config == null ) throw new Exception("Invalid configuration data");
	    
// upload the bitstream 
		String configFN = ztex.config.defaultBitstreamPath("memfifo");
		System.out.println("Found " + ztex.config.getName()+",  using bitstream "+configFN);
	        System.out.println("FPGA configuration time: " + ztex.configureFpga( configFN , true, -1 ) + " ms");
	        
// upload the bitstream 
		boolean isFx3 = ztex.dev().fx3();
	        int sizeFact = isFx3 ? 2 : 1;

// claim tnterface
    	    	ztex.trySetConfiguration ( 1 );
    	    	ztex.claimInterface ( 0 );
		ztex.debug2PrintNextLogMessages(System.out); 

// start traffic reader
		reader = new UsbReader( ztex );
		reader.start();

// Mode 1: 208 MByte/s Test data generator: used for speed test
		ztex.setMode(1);
		System.out.println((isFx3 ? "208" : "48") + " MByte/s test data generator: ");
		long t0 = new Date().getTime();
		ztex.verify(reader, 1000*sizeFact, 0, verbose);
		System.out.println("Read data rate: " + Math.round(reader.bufSize*sizeFact*1000.0/((new Date().getTime()-t0)*100.0)*0.1) + " MByte/s");
		ztex.debug2PrintNextLogMessages(System.out); 

// Mode 2: 13 MByte/s Test data generator: used for speed test
		ztex.setMode(2);
		System.out.println((isFx3 ? "13" : "12") + " MByte/s test data generator: ");
		t0 = new Date().getTime();
		ztex.verify(reader, 500*sizeFact, 0, verbose);
		ztex.debug2PrintNextLogMessages(System.out);

// Mode 0: write+read mode
		System.out.println("USB write + read mode: on certain USB implementations concurrent reading / writing may cause deadlocks or timeouts");
		writer = new UsbTestWriter( ztex );
		if ( ztex.config.getRamSize()>1024*1024 ) reader.writer = writer.writer();  // stops the reader until writer has been started
		ztex.reset();
//		ztex.setMode(3);
		writer.start();
		reader.reset();
		ztex.debug2PrintNextLogMessages(System.out); 

		System.out.println("USB write + read mode: speed test");
		t0 = new Date().getTime();
		ztex.verify(reader, 750*sizeFact, 0, verbose);
		System.out.println("Read + Write data rate: " + Math.round(reader.bufSize* sizeFact*750.0/((new Date().getTime()-t0)*100.0)*0.1) + " MByte/s");
		ztex.debug2PrintNextLogMessages(System.out); 
        	
		System.out.println("USB write + read mode: 10 MByte/s read test");
		ztex.verify(reader, 400*sizeFact, 10000, verbose);
		ztex.debug2PrintNextLogMessages(System.out); 

//		if ( writer.error() ) System.out.println( "Write errors occurred: "+writer.errorBufferCount() ); 
	    
// Terminating threads
		writer.terminate(); 		// stop the writer
		reader.terminate();  		// stop the reader
		ztex.reset();

		for (int i=0; i<30000 && (writer.isAlive() || reader.isAlive()); i++ ) { 
		    if ( ( i % 1000 ) == 999 ) {
			System.out.print("Waiting for termination of: ");
			if ( reader.isAlive() ) System.out.print("Reader ");
			if ( writer.isAlive() ) System.out.print("Writer");
			System.out.println();
		    }
		    try {
			Thread.sleep(1);
        	    }
    		    catch ( InterruptedException e) {
        	    } 
		}
		
		if ( writer.isAlive() || reader.isAlive() ) {
		    System.out.print("Threads do not want to die: ");
		    if ( reader.isAlive() ) System.out.print("Reader ");
		    if ( writer.isAlive() ) System.out.print("Writer");
	    	    System.out.println();
	    	} else {
		    System.out.println("Threads terminated.");
    		    ztex.releaseInterface( 0 );	 
		}

// releases interface	    
    	    }
	    catch (Exception e) {
		if ( writer!=null ) writer.terminate(); 	// stop the writer
		if ( reader!=null ) reader.terminate();  	// stop the reader
		System.out.println("Error: "+e.getLocalizedMessage() );
		ztex.debug2PrintNextLogMessages(System.out); 
	    } 
	}
	catch (Exception e) {
	    System.out.println("Error: "+e.getLocalizedMessage() );
	} 
	
   } 
   
}
